<?php
namespace Zeedhi\DataMining\Common\Service\DataCollector;

use Zeedhi\DataMining\Common\Service\Connection;
use Zeedhi\DataMining\Common\Util\Environment;
use Zeedhi\DataMining\Common\Util\Period as Moment;
use Zeedhi\DataMining\Common\Service\Mongo\Runner as MongoRunner;
use Zeedhi\DataMining\Common\Service\DataSource;
use Zeedhi\DataMining\Common\Service\Period;
use Zeedhi\Framework\DataSource\Manager\Mongo\Types\Type;
use Zeedhi\Framework\DataSource\FilterCriteria;
use Zeedhi\Framework\Remote\Server;
use Zeedhi\Zmart\DataSource\ZmartFilterCriteria;
use Zeedhi\Framework\DTO;

class Import {

    const ONE_SECOND_IN_MICROSECONDS = 1000000;

    const SYNC_ITEMS_NUMBER = 5000;
    const TEMP_COLLECTION_PREFFIX = 'temp_';
    const REGISTRATION_TEMPORARY_COLLECTIONS = 'registrationTempCollections';
    const PAGE_SIZE = 5000;

    const DATE_COLUMN_PROPERTY = 'dateFilterColumn';

    // Type of save routines
    const CONCAT         = 'concat';
    const DROP           = 'drop';
    const DROP_BY_PERIOD = 'dropByPeriod';

    protected $sendMail = false;

    protected $environment;
    protected $mongoRunner;
    protected $historyService;
    protected $dataSourceService;
    protected $periodService;
    protected $socketClient;
    protected $dateType;
    protected $connectionService;

    public function __construct(Environment $environment, MongoRunner $mongoRunner, History $historyService, DataSource $dataSourceService, Period $periodService,
                                SocketClient $socketClient, Connection $connectionService, Server $zmartServer) {
        $this->environment = $environment;
        $this->mongoRunner = $mongoRunner;
        $this->historyService = $historyService;
        $this->dataSourceService = $dataSourceService;
        $this->periodService = $periodService;
        $this->socketClient = $socketClient;
        $this->connectionService = $connectionService;
        $this->zmartServer = $zmartServer;
        $this->dateType = Type::getType(Type::MONGO_DATE);
    }

    public function prepareEnvironmentToSync($organizationId, $dataSourceId) {
        $this->environment->setOrganization($organizationId);
        $process = $this->generateProcessId($dataSourceId);
        $this->shuttingDownPreviousImportProcesses($dataSourceId, $process,$organizationId);
        $this->createProcessHistory($process, $organizationId, $dataSourceId);
        $this->addStatusImportationToHistory($process, 'queue');
        return $process;
    }

    public function startProcessSync($organizationId, $dataSourceId, $period = null, $persistType = null, $importType = null) {
        $process = $this->prepareEnvironmentToSync($organizationId, $dataSourceId);

        try {
            $this->updateHistory($process, array('importType' => $importType));
            $this->addProcessLog($process, History::LOG_DEBUG, "Starting {$importType} import process.");
            $dataSource = $this->dataSourceService->findById($dataSourceId);

            $this->log("Getting process informations");
            $connectionData = $this->connectionService->findByOrganization($organizationId);
            $connectionId = isset($dataSource["connection"]) ? $dataSource["connection"]: null;
            $connectionParams = $this->connectionService->getConnectionOrMain($connectionId);

            if(isset($connectionParams["room"])){
                $connectionClientRoom = $organizationId."_".$connectionParams["room"];
            } else {
                $connectionClientRoom = isset($connectionData["clientRoom"]) ? $connectionData["clientRoom"] : $organizationId;
            }

            $isJoinDataSource = (isset($dataSource['joinDataSource']) && $dataSource['joinDataSource']) ? true : false;

            $interval = $this->periodService->decodePeriod($period ?: $dataSource['period'],$this->connectionService);
            $start = $interval['start'];
            $end = $interval['end'];

            $this->addStartAndEndToHistory($process, $start, $end);

            if(!$isJoinDataSource){
                $this->startRegularSync($organizationId, $dataSource, $process, $start, $end, $persistType, $connectionClientRoom);
            } else {
                $this->startJoinSync($organizationId, $process, $dataSource, $start, $end, $persistType);
            }

            return $process;
        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
            $this->log("ERROR: ".$error->getMessage());
        }

        return $process;
    }

    protected function startRegularSync($organizationId, $dataSource, $process, $start, $end, $persistType, $connectionClientRoom){
        $connectionId = isset($dataSource["connection"]) ? $dataSource["connection"]: null;
        $connectionParams = $this->connectionService->getConnectionOrMain($connectionId);

        $connectionURLSsl = isset($connectionParams['dbUrlSsl']) ? $connectionParams['dbUrlSsl'] : null;
        if(!$connectionURLSsl && isset($connectionParams['host'])&& isset($connectionParams['port'])&& isset($connectionParams['dbname'])){
            $connectionURLSsl = OpensslCrypt::encrypt("jdbc:oracle:thin:@".$connectionParams['host'].":".$connectionParams['port']."/".$connectionParams['dbname']);
        }
        $connectionURL = isset($connectionParams['dbUrl']) ? $connectionParams['dbUrl'] : $connectionURLSsl;

        $this->addStatusImportationToHistory($process, 'start');

        $pageSize = $this->getPageSize($dataSource);
        $this->addPageSizeToHistory($process, $pageSize);

        $this->createTemporaryCollection($process, $organizationId, $dataSource, $start, $end, $persistType);

        $scriptBefore                 = isset($dataSource['beforeImportScript'])           ? $this->injectPeriodInQuery($dataSource['beforeImportScript'], $start, $end) : '';
        $scriptAfter                  = isset($dataSource['afterImportScript'])            ? $this->injectPeriodInQuery($dataSource['afterImportScript'], $start, $end)  : '';
        $execAfterIfImportError       = isset($dataSource['execAfterIfImportError'])       ? $dataSource['execAfterIfImportError']       : false;
        $execAfterIfBeforeScriptError = isset($dataSource['execAfterIfBeforeScriptError']) ? $dataSource['execAfterIfBeforeScriptError'] : false;
        $dataSource['dialect']        = isset($connectionParams['dialect'])                ? $connectionParams['dialect'] : null;
        $dataSource['driverClass']    = isset($connectionParams['driverClass'])            ? $connectionParams['driverClass'] : null;
        $passwordSsl                  = isset($connectionParams['passwordSsl'])            ? $connectionParams['passwordSsl'] : null;
        $password                     = isset($connectionParams['password'])               ? $connectionParams['password'] : null;
        $usernameSsl                  = isset($connectionParams['usernameSsl'])            ? $connectionParams['usernameSsl'] : null;
        $username                     = isset($connectionParams['username'])               ? $connectionParams['username'] : null;

        $addressee                    = isset($dataSource['addressee'])                    ? $dataSource['addressee'] : null;
        $titleEmail                   = isset($dataSource['titleEmail'])                   ? $dataSource['titleEmail'] : null;
        $textEmail                    = isset($dataSource['textEmail'])                    ? $dataSource['textEmail'] : null;
        $finality                     = isset($dataSource['finality'])                     ? $dataSource['finality'] : 'I';
        $options                      = isset($dataSource['options'])                     ? $dataSource['options'] : null;

        $this->addProcessLog($process, History::LOG_DEBUG, "Start sync");
        $this->startSync($organizationId, $process, $dataSource, $start, $end, $pageSize, $connectionURL, $username,
            $password, $scriptBefore, $scriptAfter, $execAfterIfImportError, $execAfterIfBeforeScriptError,
            $connectionClientRoom,$usernameSsl,$passwordSsl,$connectionURLSsl, $addressee, $titleEmail, $textEmail, $finality,$options);
    }

    protected function injectIdentification($query, $dataSource, $organizationId)
    {
        $name = $dataSource['name'];
        $queryIdentification = <<<SQL
/* ZmartBI - $organizationId - $name */

SQL;
        return $queryIdentification . $query;
    }


    protected function startSync($organizationId, $process, $dataSource, $start, $end, $pageSize, $dbUrl, $userName, $password, $scriptBefore, $scriptAfter,
                                 $execAfterIfImportError, $execAfterIfBeforeScriptError, $clientRoom, $usernameSsl, $passwordSsl, $connectionURLSsl,
                                 $addressee = '', $titleEmail = '', $textEmail ='', $finality ='I', $options = array())
    {

        $importHistory = $this->historyService->findByProcessId($process);
        if(isset($importHistory['importStatus']) && $importHistory['importStatus'] !== 'canceled'){
            $query = $this->injectPeriodInQuery($dataSource['query'], $start, $end);
            $query = $this->injectNrorgInQuery($query, $organizationId);
            $query = $this->injectIdentification($query,$dataSource,$organizationId);
            $queryKeys = $this->extractKeyColumnsName($dataSource);
            $dialect = isset($dataSource['dialect']) ? $dataSource['dialect'] : null;
            $driverClass = isset($dataSource['driverClass']) ? $dataSource['driverClass'] : null;

            try {
                $this->log("Start client importation");
                $this->socketClient->startImportation($organizationId, $process, $query, $queryKeys, $pageSize, $dbUrl,
                                                        $userName,  $password, $scriptBefore, $scriptAfter,
                                                        $execAfterIfImportError, $execAfterIfBeforeScriptError, $clientRoom,
                                                        $dialect, $driverClass, $usernameSsl, $passwordSsl, $connectionURLSsl,
                                                        $addressee, $titleEmail, $textEmail, $finality,$options);
            } catch (\Exception $error) {
                $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
                $this->addProcessLog($process, History::LOG_DEBUG, $error->getMessage());
                $this->log("ERROR: ".$error->getMessage());
            }
        } else {
            $this->notifyImportUpdate($process);
        }
    }

    protected function injectPeriodInQuery($query, $start, $end) {
        $query = str_ireplace(':START_DATE', "'{$start}'", $query);
        $query = str_ireplace(':END_DATE', "'{$end}'", $query);
        return $query;
    }

    protected function injectNrorgInQuery($query, $nrorg) {
        $query = str_ireplace(':NRORG', "'{$nrorg}'", $query);
        return $query;
    }

    protected function extractKeyColumnsName(array $dataSource) {
        $keys = array();
        if (!empty($dataSource['columns'])) {
            foreach ($dataSource['columns'] as $column) {
                if (isset($column['isKey']) && $column['isKey']){
                    $keys[] = $column['name'];
                }
            }
        }
        return $keys;
    }

    protected function generateProcessId($dataSourceId) {
        $this->log("Generate process id");
        return uniqid().mt_rand()."_{$dataSourceId}";
    }

    protected function getPageSize($dataSource) {
        return !empty($dataSource['pageSize']) ? $dataSource['pageSize'] : self::SYNC_ITEMS_NUMBER;
    }

    protected function createTemporaryCollection($process, $organizationId, $dataSource, $start, $end, $persistType = null) {
        $this->log("Create temporary collection");
        $dateFilterColumn = isset($dataSource[self::DATE_COLUMN_PROPERTY]) ? $dataSource[self::DATE_COLUMN_PROPERTY] : null;
        $this->updateHistory($process, array(
            'persistType' => $persistType,
            self::DATE_COLUMN_PROPERTY => $dateFilterColumn
        ));

        try {
            $tempCollection = self::TEMP_COLLECTION_PREFFIX.$process;
            $this->mongoRunner->createCollection($this->environment->getOrganizationConnection(), $tempCollection);
            $this->saveDataTempCollection($tempCollection);
            $this->addProcessLog($process, History::LOG_DEBUG, "Temporary collection created. {$tempCollection}");
        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
            $this->addProcessLog($process, History::LOG_DEBUG, $error->getMessage());
        }
    }

    protected function removeDataByPeriod($dateFilterColumn, $collection, $start, $end) {
        $start = \DateTime::createFromFormat('d/m/Y H:i:s', "{$start} 00:00:00");
        $start = $this->fixTimeZoneMongo($start);
        $start = $this->dateType->convertToDatabaseValue($start);

        $end = \DateTime::createFromFormat('d/m/Y H:i:s', "{$end} 23:59:59");
        $end = $this->fixTimeZoneMongo($end);
        $end   = $this->dateType->convertToDatabaseValue($end);

        $filterCriteria = new FilterCriteria($collection);

        if($dateFilterColumn){
            $filterCriteria->addCondition($dateFilterColumn, FilterCriteria::GTE, $start);
            $filterCriteria->addCondition($dateFilterColumn, FilterCriteria::LTE, $end);
        }

        $this->mongoRunner->remove($this->environment->getOrganizationConnection(), $filterCriteria);
    }

    protected function fixTimeZoneMongo(\DateTime $date){
        $offset = $date->getTimezone()->getOffset($date);
        $modify =  $offset.'seconds';
        return $date->modify($modify);
    }

    protected function dropCollection($collectionName) {
        $orgConnection = $this->environment->getOrganizationConnection();
        $orgConnection->dropCollection($collectionName);
    }

    protected function createProcessHistory($process, $organizationId, $dataSourceId) {
        $this->log("Create Process History");
        $this->historyService->create($process, $organizationId, $dataSourceId);
        $this->addProcessLog($process, History::LOG_DEBUG, "Process {$process} started.");
    }

    protected function addStartAndEndToHistory($process, $start, $end) {
        $this->updateHistory($process, array('start' => $start, 'end' => $end));
        $this->notifyImportUpdate($process);
        $this->addProcessLog($process, History::LOG_DEBUG, "Start and end date setted. Start: {$start}; End: {$end}");
    }

    protected function addStatusImportationToHistory($process, $importStatus) {
        $this->updateHistory($process, array('importStatus' => $importStatus));
    }

    protected function addPageSizeToHistory($process, $pageSize) {
        $this->log("Add page size to history");
        try {
            $this->updateHistory($process, array('pageSize' => $pageSize));
            $this->addProcessLog($process, History::LOG_DEBUG, "Page size configured: {$pageSize}.");
        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
            $this->addProcessLog($process, History::LOG_DEBUG, $error->getMessage());
        }
    }

    public function setErrorOnProcess($process, $message, $code = null, $trace = null) {
        $processHistory = $this->historyService->findByProcessId($process);
        $organizationId = $processHistory['organizationId'];
        $dataSourceId = empty($processHistory['dataSourceId']) ? null : $processHistory['dataSourceId'] ;

        if($processHistory['importStatus'] !== 'done') {
            $this->environment->setOrganization($organizationId);
            $dataSource = $this->dataSourceService->findById($dataSourceId);

            $this->addStatusImportationToHistory($process, 'failed');
            $this->updateHistory($process, array('hasError' => true));
            $this->setEndDateAndElapsedTimeOnProcess($process);
            $this->addProcessLog($process, History::LOG_ERROR, $message, $code, $trace);
            $this->dropCollection(self::TEMP_COLLECTION_PREFFIX.$process);

            $this->notifyImportUpdate($process);
            $this->socketClient->notifyError($this->environment->getOrganization(), $process);
            $this->notifyImportationEndToZmartBackend($organizationId, $dataSourceId, $process, true);
        } else {
            $this->addProcessLog($process, History::LOG_ERROR, "An error has occured after importation was finished with the message: {$message}");
        }

        if(isset($dataSource['createdBy']) && $this->sendMail !== $process){
            $this->socketClient->notifyErrorOfEmail($processHistory, $dataSource['createdBy'], $dataSource['name'], $message);
            $this->sendMail = $process;
        }
    }

    protected function updateHistory($process, $dataToUpdate) {
        $criteria = array('process' => $process);
        $update   = array('$set' => $dataToUpdate);
        $this->historyService->update($criteria, $update);
    }

    public function setProcessLength($process, $length) {
        try {
            $this->updateHistory($process, array('subprocesses' => $length));
            $this->addProcessLog($process, History::LOG_DEBUG, "Process length: {$length}.");

            if ($length === 0) {
                $processHistory = $this->historyService->findByProcessId($process);
                $this->environment->setOrganization($processHistory['organizationId']);

                $tempCollection = self::TEMP_COLLECTION_PREFFIX.$process;
                $this->dropCollection($tempCollection);

                $dataSourceId = empty($processHistory['dataSourceId']) ? null : $processHistory['dataSourceId'] ;
                $this->finalize($process, $dataSourceId);
            }
        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
        }
    }
    
    public function setQueryLength($process, $length)
    {
        try {
            $this->updateHistory($process, array('subprocesses' => 1));
            $this->addProcessLog($process, History::LOG_DEBUG, "Result query length: {$length}.");

        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
        }
    }

    public function sendEmailFinalize($process)
    {
        try {
            $processHistory = $this->historyService->findByProcessId($process);
            $this->environment->setOrganization($processHistory['organizationId']);

            $tempCollection = self::TEMP_COLLECTION_PREFFIX . $process;
            $this->dropCollection($tempCollection);

            $dataSourceId = empty($processHistory['dataSourceId']) ? null : $processHistory['dataSourceId'];
            $this->finalize($process, $dataSourceId);

            $this->updateHistory($process, array('imported' => 1));

        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
        }
    }

    public function sync($organizationId, $process, $page, array $items) {
        $this->environment->setOrganization($organizationId);
        $processHistory = $this->historyService->findByProcessId($process);
        $processImportStatus = isset($processHistory['importStatus']) ? $processHistory['importStatus'] : null;

        if (!$processHistory['hasError'] && $processImportStatus !== "canceled") {
            try {
                if($processImportStatus !== 'blocked'){
                    $this->addStatusImportationToHistory($process, 'blocked');
                }

                $organizationId = $processHistory['organizationId'];
                $dataSourceId   = empty($processHistory['dataSourceId']) ? null : $processHistory['dataSourceId'];

                $this->addStatusImportationToHistory($process, 'importing');

                $this->environment->setOrganization($organizationId);
                $this->saveItems($process, $items, $dataSourceId, $organizationId);

                $this->getProcessRowLock($process);

                $this->incrementImportedNumber($process, $page);
                $this->notifyImportUpdate($process);
                $this->checkIfIsTheLastSubprocess($process, $dataSourceId);
            } catch (\Exception $error) {
                $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
            }

            $this->removeProcessRowLock($process);
        } else {
            $this->notifyImportUpdate($process);
        }
    }

    protected function getProcessRowLock($process) {
        $connection = $this->environment->getOrganizationConnection();

        $criteria = array(
            'process' => $process,
            'locked' => false
        );

        $update = array(
            '$set' => array(
                'locked' => true
            )
        );

        $testCount = 0;

        do {
            $updateStatus = $this->mongoRunner->update(
                $connection,
                History::COLLECTION_NAME,
                $criteria,
                $update
            );

            $lockAquired = $updateStatus->getModifiedCount() === 1;

            if (!$lockAquired) {
                if ($testCount === 200) {
                    throw new \Exception('Process locked for too long');
                }

                usleep(self::ONE_SECOND_IN_MICROSECONDS / 2);
            }

            $testCount++;
        } while(!$lockAquired);
    }

    protected function removeProcessRowLock($process) {
        $criteria = array(
            'process' => $process
        );

        $update = array(
            '$set' => array(
                'locked' => false
            )
        );

        $this->mongoRunner->update(
            $this->environment->getOrganizationConnection(),
            History::COLLECTION_NAME,
            $criteria,
            $update
        );

    }

    protected function notifyImportUpdate($process){
        $processHistory = $this->historyService->findByProcessId($process);
        $organizationId = $processHistory['organizationId'];
        $this->socketClient->notifyImportUpdate($organizationId, $processHistory);
    }

    public function addProcessLog($process, $logType = History::LOG_DEBUG, $logMessage) {
        $this->notifyImportUpdate($process);
        $this->historyService->addLog($process, $logType, $logMessage);
    }

    protected function saveItems($process, array $items, $dataSourceId, $organizationId) {
        $orgConnection  = $this->environment->getOrganizationConnection();

        if ($dataSourceId) {
            $tempCollection = self::TEMP_COLLECTION_PREFFIX . $process;
            $dataSource = $this->dataSourceService->findById($dataSourceId);
            $this->convertDateValues($items, $dataSource);
        } else {
            $tempCollection = $process;
        }
        return $orgConnection->insertMany($tempCollection, $items);
    }

    protected function convertDateValues(array &$items, $dataSource) {
        foreach ($items as &$item) {
            $this->convertDateValue($item, $dataSource);
        }
    }

    protected function convertDateValue(&$item, $dataSource) {
        foreach($dataSource['columns'] as $column) {
            $columnName = $column['name'];
            if (isset($column['isDate']) && $column['isDate'] && isset($item[$columnName]) && $item[$columnName]) {
                $date = $this->dateType->convertToDatabaseValue($item[$columnName]);
                $item[$columnName] = $date ?: $item[$columnName];
            }
        }
    }

    protected function incrementImportedNumber($process, $page) {
        $criteria = array('process' => $process);
        $update   = array('$inc' => array('imported' => 1));
        $this->historyService->update($criteria, $update);

        $processHistory = $this->historyService->findByProcessId($process);
        $this->addProcessLog($process, History::LOG_DEBUG, "Finish subprocess {$processHistory['imported']} of {$processHistory['subprocesses']}, page {$page}.");
    }

    protected function checkIfIsTheLastSubprocess($process, $dataSourceId) {
        $importHistory = $this->historyService->findByProcessId($process);
        if ($importHistory['imported'] === $importHistory['subprocesses']) {
            $this->addProcessLog($process, History::LOG_DEBUG, "Start data persistence");
            $this->replaceOriginalCollection($process, $dataSourceId, $importHistory);
            $this->finalize($process, $dataSourceId);
        }
    }

    protected function finalize($process, $dataSourceId){
        if ($dataSourceId) {
            $this->checkForDependentDataSources($dataSourceId, $process);
        }
        $this->addProcessLog($process, History::LOG_DEBUG, "Finalize process");
        $this->addStatusImportationToHistory($process, 'done');
        $endDate = $this->setEndDateAndElapsedTimeOnProcess($process);
        $this->addLastUpdateOnDataSource($dataSourceId, $endDate, $process);
        $this->notifyImportationEnd($dataSourceId, $process);
    }

    protected function getDataSourceInfo($dataSourceID) {
        $conn = $this->environment->getOrganizationConnection();
        return $this->mongoRunner->findById(
            $conn,
            'dataSource',
            new \MongoDB\BSON\ObjectId($dataSourceID)
        );
    }

    protected function notifyImportationEnd($dataSourceId, $process) {
        $organizationId = $this->environment->getOrganization();
        if ($dataSourceId) {
            $dataSource = $this->getDataSourceInfo($dataSourceId);
            $updateIndicators = isset($dataSource['willUpdateIndicators']) && $dataSource['willUpdateIndicators'];
        } else {
            $dataSourceId = $process;
            $updateIndicators = true;
        }
        $this->socketClient->notifyImportationEnd($organizationId, $dataSourceId, $updateIndicators);
        $this->notifyImportationEndToZmartBackend($organizationId, $dataSourceId, $process);
    }

    protected function notifyImportationEndToZmartBackend($organizationId, $dataSourceId,
                                                          $process, $hasError = false) {
        $row = array(
            'organizationId' => $organizationId,
            'dataSourceId' => $dataSourceId,
            'processId' => $process,
            'importSucceeded' => !$hasError
        );

        $method = 'POST';
        $routePath = '/finishImportation';
        $request = new DTO\Request\Row($row, $method, $routePath, 'zmart');
		$this->zmartServer->request($request);
    }

    protected function replaceOriginalCollection($process, $dataSourceId, $importHistory) {
        try {
            $dateFilterColumn = isset($importHistory['dateFilterColumn']) ? $importHistory['dateFilterColumn'] : null;
            $persistType = isset($importHistory['persistType']) ? $importHistory['persistType'] : null;

            if($persistType === 'drop'){
                $this->addProcessLog($process, History::LOG_DEBUG, "Drop old records");
                $this->dropCollection($dataSourceId);
            } else if($persistType === 'dropByPeriod'){
                $this->addProcessLog($process, History::LOG_DEBUG, "Drop old records by period");
                $start = isset($importHistory['start']) ? $importHistory['start'] : null;
                $end = isset($importHistory['end']) ? $importHistory['end'] : null;
                $this->removeDataByPeriod($dateFilterColumn, $dataSourceId, $start, $end);
            }

            $mongoConnection = $this->environment->getOrganizationConnection();
            $tempCollection = self::TEMP_COLLECTION_PREFFIX.$process;
            $this->addProcessLog($process, History::LOG_DEBUG, "Update original collection");
            $this->copyDocumentsToCollection($tempCollection, $dataSourceId, $process);
            $this->addProcessLog($process, History::LOG_DEBUG, "Original collection updated");
            $this->dropCollection($tempCollection);
        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
        }
    }

    protected function copyDocumentsToCollection($originCollection, $destinCollection, $process) {
        try {
            $this->addStatusImportationToHistory($process, 'saving');
            $mongoConnection = $this->environment->getOrganizationConnection();
            $page = 0;
            $rows = array();
            do {
                $options = array(
                    "sort" => array("_id" => 1),
                    "skip" => $page * self::PAGE_SIZE,
                    "limit" => self::PAGE_SIZE
                );
                $cursor = $mongoConnection->getCollection($originCollection)->find([],$options);
        		$rows = iterator_to_array($cursor);
                $page++;

                if(count($rows) !== 0) {
                    $mongoConnection->insertMany($destinCollection, $rows);
                }
            } while(count($rows) !== 0);
        } catch (\Exception $error) {
            throw $error;
        }
    }

    protected function setEndDateAndElapsedTimeOnProcess($process) {
        $processHistory = $this->historyService->findByProcessId($process);

        $createDate = $this->dateType->convertToPHPValue($processHistory['createDate']);
        $endDate = new \DateTime();
        $elapsedTime = $endDate->getTimestamp() - $createDate->getTimestamp();

        $endDate = $this->dateType->convertToDatabaseValue($endDate);
        $criteria = array('_id' => $processHistory['_id']);
        $update = array(
            '$set' => array(
                'endDate'     => $endDate,
                'elapsedTime' => $elapsedTime
            )
        );

        $this->historyService->update($criteria, $update);
        $this->notifyImportUpdate($process);
        $this->addProcessLog($process, History::LOG_DEBUG, "Importation process done. Elapsed time: {$elapsedTime} seconds.");
        return $endDate;
    }

    protected function addLastUpdateOnDataSource($dataSourceId, $endDate, $process) {
        if ($dataSourceId) {
            $mongoConnection = $this->environment->getOrganizationConnection();
            $collectionName = 'dataSource';
            $criteria = array('_id' => new \MongoDB\BSON\ObjectId($dataSourceId));
            $update = array('$set' => array('lastUpdate' => $endDate));
            $mongoConnection->update($collectionName, $criteria, $update);
        }
    }

    public function removeTempDataSourceCollections($organizationId, $dataSourceId){
        $this->log("Remove temporary collections of dataSource");
        $this->environment->setOrganization($organizationId);
        $orgConnection = $this->environment->getOrganizationConnection();

        $collectionsNames = $orgConnection->getCollectionNames();
        foreach ($collectionsNames as $collectionName) {
            $isDataSourceTemp = strpos($collectionName, $dataSourceId);

            if($isDataSourceTemp && (substr($collectionName, 0, 5) === self::TEMP_COLLECTION_PREFFIX)) {
                $orgConnection->dropCollection($collectionName);
            }
        }
    }

    //Join DataSources

    protected function finalizeJoinProcess($process, $dataSource, $items, $start, $end, $persistType){
        try {
            $processHistory = $this->historyService->findByProcessId($process);

            if(isset($processHistory["importStatus"]) && $processHistory["importStatus"] !== "canceled"){
                $this->addStatusImportationToHistory($process, 'blocked');
                $dateFilterColumn = isset($dataSource[self::DATE_COLUMN_PROPERTY]) ? $dataSource[self::DATE_COLUMN_PROPERTY] : null;
                $this->addProcessLog($process, History::LOG_DEBUG, "Start persist {$persistType}");
                $collectionName = $dataSource['_id'];

                $orgConnection = $this->environment->getOrganizationConnection();

                if($persistType){
                    switch ($persistType) {
                        case Import::DROP_BY_PERIOD:
                            $this->removeDataByPeriod($dateFilterColumn, $collectionName, $start, $end);
                            break;
                        case Import::DROP:
                            $orgConnection->dropCollection($collectionName);
                            break;
                    }
                }

                $orgConnection->insertMany($collectionName, $items);
                $this->notifyImportUpdate($process);
                $this->finalize($process, $collectionName);
            } else {
                $this->notifyImportUpdate($process);
            }
        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
        }
    }

    public function noDependentDataSourcesToUpdate($processId){
        $this->addProcessLog($processId, History::LOG_DEBUG, "No dependent datasources");
    }

    protected function checkForDependentDataSources($originDataSourceId, $processId){
        $organizationId = intval($this->environment->getOrganization());
        $filterCriteria = new FilterCriteria('dependentDataSources');
        $filterCriteria->addCondition('organizationId', $organizationId);
        $filterCriteria->addCondition('originDataSourceId', $originDataSourceId);

        $data = $this->mongoRunner->findBy($this->environment->getZmartConnection(), $filterCriteria);

        if (isset($data) && $data){
            $data = reset($data);
            $dataSourcesToUpdate = $data['dependentDataSources'];

            foreach ($dataSourcesToUpdate as $dataSourceToUpdate) {
                $this->syncJoinDataSource($organizationId, $dataSourceToUpdate, $processId);
            }
        } else {
            $this->noDependentDataSourcesToUpdate($processId);
        }
    }

    public function syncJoinDataSource($organizationId, $joinDataSourceId, $processId){
        $this->addProcessLog($processId, History::LOG_DEBUG, "Start updating dependent datasource $joinDataSourceId");
        $this->prepareEnvironmentToSync($organizationId, $joinDataSourceId);
    }

    protected function getGeneratedFacts($joinDataSource, $process){
        $this->addProcessLog($process, History::LOG_DEBUG, "Get generated facts from join datasource");
        $facts = array();

        foreach ($joinDataSource['columns'] as $column) {
            if($column['isGenerated'] && $column['isFact']){
                array_push($facts, $column['name']);
            }
        }

        return $facts;
    }

    protected function startJoinSync($organizationId, $process, $dataSource, $start, $end, $persistType = null){
        try {
            $mongoConnection = $this->environment->getOrganizationConnection();
            $dataSource['dataSourcesInformations'] = self::objectToArray($dataSource['dataSourcesInformations']);
            $keys = $dataSource['keys'];
            $facts = $this->getGeneratedFacts($dataSource, $process);

            $this->addStatusImportationToHistory($process, 'start');

            $originDataSources = $this->getOriginDataSources($dataSource, $mongoConnection, $start, $end, $process);
            $newDataSource = $this->generateNewDataSource($originDataSources, $keys, $facts, $dataSource, $process);
            $newData = $this->getAggregatedFacts($newDataSource, $dataSource, $process);

            $this->finalizeJoinProcess($process, $dataSource, $newData, $start, $end, $persistType);
        } catch (\Exception $error){
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
            $this->addProcessLog($process, History::LOG_DEBUG, $error->getMessage());
        }
    }

    protected function convertStringToMongoDate($date, $startOrEnd) {
        $momentDate = Moment::convertDateTimeToMoment(\DateTime::createFromFormat(Moment::DATE_FORMAT, $date));
        if($startOrEnd =='start') {
            $momentDate->startOf(Moment::UNIT_DAY);
        } else if($startOrEnd =='end') {
            $momentDate->endOf(Moment::UNIT_DAY);
        }
        $mongoDate = $this->dateType->convertToDatabaseValue($momentDate);
        return $mongoDate;
    }

    public function getOriginDataSources($joinDataSource, $mongoConnection, $start, $end, $process){
        $joinDataSourceInformations = $joinDataSource['dataSourcesInformations'];
        $originDataSources = array();

        foreach ($joinDataSourceInformations as $originDataSourceId => $originDataSourceInformation) {
            $this->addProcessLog($process, History::LOG_DEBUG, "Get origin datasource informations from {$originDataSourceId}");
            $dimensions = array();

            if(!empty($originDataSourceInformation['keysTranslation'])){
                foreach ($originDataSourceInformation['keysTranslation'] as $keyName => $keyTranslation) {
                    array_push($dimensions, $keyName);
                }
            }

            $filterCriteria = new ZmartFilterCriteria($originDataSourceId);
            $filterCriteria->setGroupBy($dimensions);

            $dateFilterColumn = $originDataSourceInformation['dateFilterColumn'];
            $this->setFilterCriteriaDateParams($filterCriteria, $dateFilterColumn, $start, $end);

            foreach ($originDataSourceInformation['facts'] as $fact => $aggFunction) {
                $filterCriteria->addAggregateFunction($fact, $aggFunction, $fact);
            }

            $dataSource = $this->mongoRunner->aggregate($mongoConnection, $filterCriteria, $originDataSourceId);
            $originDataSources[$originDataSourceId] = $dataSource;
        }

        return $originDataSources;
    }

    protected function setFilterCriteriaDateParams(&$filterCriteria, $dateFilterColumn, $start, $end) {
        if($start && $end) {
            $dates = array($this->convertStringToMongoDate($start, 'start'), $this->convertStringToMongoDate($end, 'end'));
            $filterCriteria->addCondition($dateFilterColumn, FilterCriteria::BETWEEN, $dates);
        } else if($start) {
            $filterCriteria->addCondition($dateFilterColumn, FilterCriteria::GTE, $this->convertStringToMongoDate($start, 'start'));
        } else if ($end) {
            $filterCriteria->addCondition($dateFilterColumn, FilterCriteria::LTE, $this->convertStringToMongoDate($end, 'end'));
        }
    }

    protected function generateNewDataSource($originDataSources, $keys, $facts, $dataSource, $process){
        $this->addProcessLog($process, History::LOG_DEBUG, "Start creation of join datasource");
        $this->addStatusImportationToHistory($process, 'importing');
        $newDataSource = array();

        foreach ($originDataSources as $originDataSourceId => $originDataSource) {
            $dataSourceInformation = $dataSource['dataSourcesInformations'][$originDataSourceId];

            foreach ($originDataSource as $originRow) {
                $keyValues = $this->getKeyValues($process,$originRow, $keys, $dataSourceInformation['keysTranslation']);
                $stringKeys = json_encode($keyValues);
                if(!isset($newDataSource[$stringKeys])) {
                    $newRow = $this->generateRow($process,$keyValues, $facts);
                } else {
                    $newRow = $newDataSource[$stringKeys];
                }

                $newRow = $this->translateFacts($process,$originRow, $newRow, $dataSourceInformation['factsTranslation']);

                $newDataSource[$stringKeys] = $newRow;
            }
        }
        return $newDataSource;
    }
    public static function objectToArray($object){
        if(is_object($object) && (
            'MongoDB\Model\BSONDocument' == get_class($object) ||
            'MongoDB\Model\BSONArray' == get_class($object))
        ){
            $return = iterator_to_array($object);
            foreach ($return as $key => $row){
                if(is_object($row)){
                    $return[$key] = self::objectToArray($row);
                }
            }
        }else{
            $return = $object;
        }
        return $return;
    }

    protected function getKeyValues($process,$row, $keys, $keysTranslation){
        try {
            $row["_id"] = self::objectToArray($row["_id"]);
            $newKey = array();
            foreach ($keys as $key) {
                $newKey[$key] = null;
            }

            if(is_array($row["_id"])) {
                foreach ($keysTranslation as $originalName => $newNames) {
                    if(isset($row["_id"][$originalName])) {
                        foreach ($newNames as $newName) {
                            $newKey[$newName] = $row["_id"][$originalName];
                        }
                    }
                }
                array_merge($row, $row["_id"]);
            } else {
                $key = array_shift($keysTranslation);
                foreach ($key as $newName) {
                    $newKey[$newName] = $row["_id"];
                }
            }

            unset($row["_id"]);
            return $newKey;
        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
        }
    }

    protected function generateRow($process,$keys, $facts){
        try {
            foreach ($facts as $fact) {
                $keys[$fact] = array();
            }
            return $keys;
        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
        }
    }

    protected function translateFacts($process,$originRow, $newRow, $factTranslation){
        try {
            foreach ($factTranslation as $originalName => $newNames) {
                if(isset($originRow[$originalName])) {
                    foreach ($newNames as $newName) {
                        $newRow[$newName][] = $originRow[$originalName];
                    }
                }
            }
            return $newRow;
        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
        }
    }

    protected function getFactsAggFunctionsFromColumns($joinDataSource){
        $joinAggInformations = array();

        foreach ($joinDataSource['columns'] as $column) {
            if($column['isGenerated'] && $column['isFact']){
                $joinAggInformations[$column['name']] = isset($column['joinFunction']) ? $column['joinFunction'] : $column['function'];
            }
        }

        return $joinAggInformations;
    }

    protected function getAggregatedFacts($newDataSource, $joinDataSource, $process){
        try {
            $this->addProcessLog($process, History::LOG_DEBUG, "Get aggregated facts");
            $numericArray = array();
            $joinInformations = $this->getFactsAggFunctionsFromColumns($joinDataSource);

            foreach ($newDataSource as $row) {
                foreach ($joinInformations as $fact => $function) {
                    $row[$fact] = $this->aggregateFunction($row[$fact], $function);
                }
                $numericArray[] = $row;
            }

            return $numericArray;
        } catch (\Exception $error) {
            $this->setErrorOnProcess($process, $error->getMessage(), $error->getCode(), $error->getTraceAsString());
        }
    }

    protected function aggregateFunction($fact, $joinFunction){
        $aggregatedValues = null;

        if(!empty($fact)){
            switch($joinFunction) {
                case 'MIN':
                    $aggregatedValues = min($fact);
                    break;
                case 'MAX':
                    $aggregatedValues = max($fact);
                    break;
                case 'SUM':
                    $aggregatedValues = array_sum($fact);
                    break;
                case 'COUNT':
                    $aggregatedValues = count($fact);
                    break;
                case 'AVG':
                    $aggregatedValues = array_sum($fact) / count($fact);
                    break;
            }
        }

        return $aggregatedValues;
    }

    protected function shuttingDownPreviousImportProcesses($dataSourceId, $process, $organizationId) {
        $this->log("Shutting down previous import processes");
        $history =  $this->historyService->findProcessInProgress($dataSourceId,$organizationId);
        if(!empty($history)) {
            $this->historyService->setImportStatus($history,'failed');
            if(!empty($history[0]["process"])){
                $tempCollection = self::TEMP_COLLECTION_PREFFIX.$history[0]["process"];
                if(isset($tempCollection)){
                    $this->dropCollection($tempCollection);
                }
            }
        }
    }

    protected function log($message) {
		echo "# - ".date('d/m/Y - H:i:s', time())." - ".$message.PHP_EOL;
	}

	protected function saveDataTempCollection($tempCollection) {
        $dataTempCollection = array(
            'name' => $tempCollection,
            'dateCreate' => date('Y-m-d')
        );
        $this->mongoRunner->persist($this->environment->getOrganizationConnection(), self::REGISTRATION_TEMPORARY_COLLECTIONS, $dataTempCollection);
    }

}
