<?php
namespace Zeedhi\DataMining\Common\Service;

use Moment\Moment;
use Cron\CronExpression;
use Zeedhi\Framework\DataSource\FilterCriteria;
use Zeedhi\Framework\DependencyInjection\InstanceManager;

class Period {

    const PREFIX_THIS       = 'this';
    const PREFIX_NEXT       = 'next';
    const PREFIX_LAST       = 'last';
    public static $PREFIXES = array(self::PREFIX_THIS, self::PREFIX_NEXT, self::PREFIX_LAST);

    const SUFFIX_INCLUSIVE  = 'inclusive';
    public static $SUFFIXES = array(self::SUFFIX_INCLUSIVE);

    const UNIT_DAY       = "day";
    const UNIT_WEEK      = "week";
    const UNIT_MONTH     = "month";
    const UNIT_YEAR      = "year";
    public static $UNITS = array(self::UNIT_DAY, self::UNIT_WEEK, self::UNIT_MONTH, self::UNIT_YEAR);

    const PREPOSITION_IN    = 'in';
    const PREPOSITION_UNTIL = 'until';
    public static $PREPOSITIONS = array(self::PREPOSITION_IN, self::PREPOSITION_UNTIL);

    const DATE_FORMAT = 'd/m/Y';

    public static $EXCEPTION_UNITS = array(
        'days'   => self::UNIT_DAY,
        'weeks'  => self::UNIT_WEEK,
        'months' => self::UNIT_MONTH,
        'years'  => self::UNIT_YEAR
    );

    public static $ONE_WORD_EXPRESSIONS = array(
        'today' 	=> 'this-day',
        'yesterday'	=> 'last-day',
        'tomorrow'	=> 'next-day'
    );

    protected static function getMomentNow(){
        return new Moment();
    }

    public function getNextRunDateFromExpression($expression) {
        $cron = CronExpression::factory($expression);
        return $cron->getNextRunDate();
    }

    protected function formatIntervalToReturn($start, $end){
        return array(
            'start' => $start,
            'end' => $end
        );
    }

    public function decodePeriod($period,Connection $connectionService = null) {
        if (!is_string($period)) {
            return $this->formatIntervalToReturn($period['start'], $period['end']);
        } else {
            $start = self::getMomentNow();
            $end = self::getMomentNow();

            try {
                if ($connectionService && preg_match('/^[0-9a-f]{24}$/i', $period) === 1) {
                    $periodMongo = $connectionService->findPeriodById($period);
                }
                if (!empty($periodMongo)) {
                    list($start, $end) = self::getPeriodMathExpression($periodMongo);

                } else {
                    $periodInformations = self::getPeriodParams($period);
                    $specialCase = self::checkSpecialCase($period);

                    foreach ($periodInformations as $params) {
                        $period = self::buildInterval($params, $start, $end, $specialCase);
                        $start = $period['start'];
                        $end = $period['end'];
                    }
                }

                return $this->formatIntervalToReturn($start->format('d/m/Y'), $end->format('d/m/Y'));
            } catch (\Exception $e) {
                var_dump($e->getMessage());
            }
        }
    }

    /**
     * @param $period
     * @return array
     */
    protected static function getPeriodMathExpression($period): array
    {
        $startDateExpression = !empty($period['START_DATE_EXPRESSION'])?$period['START_DATE_EXPRESSION']:'';
        $endDateExpression = !empty($period['END_DATE_EXPRESSION'])?$period['END_DATE_EXPRESSION']:'';

        $start = PeriodMathExpression\Period::getDateMoment($startDateExpression);
        $end = PeriodMathExpression\Period::getDateMoment($endDateExpression);
        return array($start, $end);
    }

    protected static function getPeriodParams($period){
        $periodInformations = array();
        $expressions = explode('|', $period);
        foreach ($expressions as $expression) {
            $param = self::decodeExpression($expression);
            array_push($periodInformations, $param);
        }

        return $periodInformations;
    }

    protected static function decodeExpression($expression){
        $splitedExpression = array();
        $expressionTerms = explode('-', $expression);
        foreach ($expressionTerms as $terms) {
            if (array_key_exists($terms, self::$ONE_WORD_EXPRESSIONS)) {
                $terms = self::$ONE_WORD_EXPRESSIONS[$terms];
            }
            $terms = explode('-', $terms);

            foreach ($terms as $term) {
                array_push($splitedExpression, $term);
            }
        }

        $decodedExpression = self::recognizeExpressionTerms($splitedExpression);
        return $decodedExpression;
    }

    protected static function recognizeExpressionTerms($splitedExpression){
        $decodedExpression = array();

        foreach ($splitedExpression as $term) {
            if (array_key_exists($term, self::$EXCEPTION_UNITS)) {
                $term = self::$EXCEPTION_UNITS[$term];
            }

            if(in_array($term, self::$PREPOSITIONS)){
                $decodedExpression['preposition'] = $term;
            } else if(in_array($term, self::$PREFIXES)){
                $decodedExpression['prefix'] = $term;
            } else if(in_array($term, self::$UNITS)){
                $decodedExpression['unit'] = $term;
            } else if(in_array($term, self::$SUFFIXES)){
                $decodedExpression['suffix'] = $term;
            } else if(!is_nan($term)){
                $decodedExpression['amount'] = $term;
            }
        }

        return $decodedExpression;
    }

    protected static function checkSpecialCase($periodCode){
        $now = self::getMomentNow();
		if((strpos($periodCode, "until-last-day") || strpos($periodCode, "until-yesterday")) && $now->getDay() === "01"){
			return true;
		}
	}

	protected static function buildInterval($params, $start, $end, $specialCase){
        $preposition = isset($params['preposition']) ? $params['preposition'] : null;
        $prefix      = isset($params['prefix'])      ? $params['prefix']      : null;
        $unit        = isset($params['unit'])        ? $params['unit']        : null;
        $suffix      = isset($params['suffix'])      ? $params['suffix']      : null;
        $amount      = isset($params['amount'])      ? $params['amount']      : 1;

        switch ($prefix) {
            case self::PREFIX_LAST:
                return self::buildIntervalToLastPrefix($preposition, $amount, $unit, $suffix, $start, $end, $specialCase);
            case self::PREFIX_NEXT:
                return self::buildIntervalToNextPrefix($preposition, $amount, $unit, $suffix, $start, $end);
            case self::PREFIX_THIS:
                return self::buildIntervalToThisPrefix($preposition, $unit, $suffix, $start, $end, $specialCase);
            case null:
                return self::buildIntervalToNoPrefix($preposition, $amount, $unit, $suffix, $start, $end);
        }
    }

    protected static function buildIntervalToLastPrefix($preposition, $amount, $unit, $suffix, $periodStart, $periodEnd, $specialCase){
        $endAmount = $suffix === self::SUFFIX_INCLUSIVE ? 0 : 1;
        $startAmount = $suffix === self::SUFFIX_INCLUSIVE ? $amount - 1 : $amount;

        $start = self::getMomentNow();
        self::subtractDate($start, $unit, $startAmount);
        $end = self::getMomentNow();
        self::subtractDate($end, $unit, $endAmount);

        self::buildIntervalsToSpecialCase($start, $end, $specialCase, $preposition);

        if($preposition === self::PREPOSITION_IN) {
            $periodStart = self::getMomentWhenInClause($periodStart, $unit, $start);
            $periodEnd = self::getMomentWhenInClause($periodEnd, $unit, $end);
        } else {
            $periodEnd = $end->endOf($unit);
            if(!$preposition){
                $periodStart = $start->startOf($unit);
            }
        }

        $interval = self::treatInterval($periodStart, $periodEnd, $unit);
        return $interval;
    }

    protected static function buildIntervalsToSpecialCase(&$start, &$end, $specialCase, $preposition){
        if($specialCase && $preposition !== self::PREPOSITION_UNTIL){
			self::subtractDate($start, self::UNIT_MONTH, 1);
			self::subtractDate($end, self::UNIT_MONTH, 1);
		}
    }

    protected static function buildIntervalToNextPrefix($preposition, $amount, $unit, $suffix, $periodStart, $periodEnd){
        $startAmount = $suffix === self::SUFFIX_INCLUSIVE ? 0 : 1;
		$endAmount   = $suffix === self::SUFFIX_INCLUSIVE ? $amount - 1 : $amount;

		$start = self::getMomentNow();
		self::addDate($start, $unit, $startAmount);
		$end = self::getMomentNow();
		self::addDate($end, $unit, $endAmount);

		if (!$preposition) {
			$periodStart = $start->startOf($unit);
			$periodEnd   = $end->endOf($unit);
		}
		else if ($preposition === self::PREPOSITION_UNTIL){
		    $periodEnd = $end->startOf($unit);
		}
		else if ($preposition === self::PREPOSITION_IN) {
			$periodStart = self::getMomentWhenInClause($periodStart, $unit, $start);
			$periodEnd = self::getMomentWhenInClause($periodEnd, $unit, $end);
		}

		$interval = self::treatInterval($periodStart, $periodEnd, $unit);
        return $interval;
    }

    protected static function getMomentWhenInClause($moment, $unit, $newMoment){
        if($unit === self::UNIT_DAY){
            $moment = $newMoment;
        }
        else if ($unit === self::UNIT_MONTH){
            $moment->setMonth($newMoment->getMonth());
            $moment->setYear($newMoment->getYear());
        }
        else if ($unit === self::UNIT_YEAR){
            $moment->setYear($newMoment->getYear());
        }

        return $moment;
    }

    protected static function treatInterval(Moment $start, Moment $end, $unit) {
        if ($unit === self::UNIT_WEEK) {
            $start->subtractDays();
            $end->subtractDays();
        }

        return array(
            'start' => $start,
            'end' => $end
        );
    }

    protected static function buildIntervalToThisPrefix($preposition, $unit, $suffix, $periodStart, $periodEnd, $specialCase){
        $start = self::getMomentNow();
        $end   = self::getMomentNow();

        self::buildIntervalsToSpecialCase($start, $end, $specialCase, $preposition);

		if (!$preposition) {
			$periodStart = $start->startOf($unit);
			$periodEnd = $end->endOf($unit);
		}
		else if ($preposition === self::PREPOSITION_UNTIL) {
			if($suffix !== self::SUFFIX_INCLUSIVE){
				self::subtractDate($end, $unit, 1);
			}
			$periodEnd = $end->endOf($unit);
		}
		else if ($preposition === self::PREPOSITION_IN) {
			$periodStart = self::getMomentWhenInClause($periodStart, $unit, $start);
			$periodEnd   = self::getMomentWhenInClause($periodEnd, $unit, $end);
		}

		$interval = self::treatInterval($periodStart, $periodEnd, $unit);
        return $interval;
    }

    protected static function buildIntervalToNoPrefix($preposition, $amount, $unit, $suffix, $periodStart, $periodEnd){
        if(!$preposition){
            $periodStart = self::setDateParam($unit, $amount, $periodStart);
        } else if($preposition === self::PREPOSITION_UNTIL){
            $periodEnd = self::setDateParam($unit, $amount, $periodEnd);
        }

        $interval = self::treatInterval($periodStart, $periodEnd, $unit);
        return $interval;
    }

    protected static function setDateParam($unit, $amount, $moment) {
		switch($unit){
			case self::UNIT_DAY:
				return $moment->setDay($amount);
			case self::UNIT_WEEK:
				return $moment->setWeek($amount);
			case self::UNIT_MONTH:
				return $moment->setMonth($amount);
			case self::UNIT_YEAR:
				return $moment->setYear($amount);
		}
	}

    public static function subtractDate(Moment &$date, $unit, $qtd = 1) {
        switch ($unit) {
            case self::UNIT_DAY:
            $date->subtractDays($qtd);
            break;
            case self::UNIT_WEEK:
            $date->subtractWeeks($qtd);
            break;
            case self::UNIT_MONTH:
            $date->subtractMonths($qtd);
            break;
            case self::UNIT_YEAR:
            $date->subtractYears($qtd);
            break;
        }
    }

    public static function addDate(Moment &$date, $unit, $qtd = 1) {
        switch ($unit) {
            case self::UNIT_DAY:
            $date->addDays($qtd);
            break;
            case self::UNIT_WEEK:
            $date->addWeeks($qtd);
            break;
            case self::UNIT_MONTH:
            $date->addMonths($qtd);
            break;
            case self::UNIT_YEAR:
            $date->addYears($qtd);
            break;
        }
    }
}